// #define __EA64__
#include "allins.hpp"
#include "bytes.hpp"
#include "detours.h"
#include "llong.hpp"
#include "logger.hpp"
#include <Zydis/Zydis.h>
#include <ida.hpp>
#include <ua.hpp>
#include <vector>
#include <windows.h>

struct insn_replacement {
  std::vector<uint8> matchbytes;
  size_t (*replace)(insn_replacement const &r, insn_t *ins);
  char const *comment = nullptr;
};

void setnop(insn_t *ins, size_t size) {
  ins->itype = NN_nop;
  ins->size = size;
}

size_t nop(insn_replacement const &r, insn_t *ins) {
  setnop(ins, r.matchbytes.size());
  return ins->size;
}

/* this is specific to the sample i had, maybe i'll cover it sometime later
   it's a bug that causes hex-rays to fail, the disassembler is not affected by
   this.
*/
std::vector<insn_replacement> const replacement_table{
  // technically, this is not correct because this instruction does something
  // however, it breaks the decompiler, and in the sample it's effectively a no-op
  // i'll get back to properly fixing this another time
  {{0xC4, 0xE2, 0xC0, 0xF7, 0xC0}, nop, "bextr bug" /*mark it in the disassembly to clear up potential confusion*/ }, // bextr bug
};

/// "Analyze one instruction and fill 'out' structure,
/// don't change 'flags', don't make changes to the database.
// returns len(instruction) in bytes, 0 if instruction can't be decoded.
void *ana_insn;
uint64 __fastcall ana_insn_replacement(void *module_data, insn_t *out,
                                       __int64 unk) {
  uint8 instr_bytes[16];
  auto read = get_bytes(instr_bytes, 16, out->ea);

  // replace disassembling behavior for stuff in |replacement_table|
  for (auto &r : replacement_table) {
    if (read < r.matchbytes.size())
      continue;

    if (memcmp(instr_bytes, r.matchbytes.data(), r.matchbytes.size()) == 0) {
      LOG("matched replacement @ %llx\n", out->ea);
      if (r.comment)
        set_cmt(out->ea, r.comment, false);

      return r.replace(r, out);
    }
  }

  // backup the instruction prior to fields being set from the orignal function
  insn_t prefill_instruction = *out;

  auto res = decltype(&ana_insn_replacement)(ana_insn)(module_data, out, unk);

  // if IDA's disassembler fails, fall back to zydis
  if (res == 0) {
    LOG("disasm fail %llx\n", out->ea);

    ZydisDisassembledInstruction zy_ins;
    if (ZYAN_SUCCESS(ZydisDisassembleIntel(ZYDIS_MACHINE_MODE_LONG_64, out->ea,
                                           instr_bytes, read, &zy_ins))) {
      // set a comment in any case
      set_cmt(out->ea, zy_ins.text, true);

      switch (zy_ins.info.mnemonic) {
      case ZYDIS_MNEMONIC_NOP: {
        *out = prefill_instruction;
        setnop(out, zy_ins.info.length);
        return zy_ins.info.length;
      }

        // handle more?

      default:
        break;
      }
    }
  }
  return res;
}

BOOL WINAPI DllMain(HINSTANCE hInstDll, DWORD dwReason, LPVOID lpReserved) {
  if (dwReason != DLL_PROCESS_ATTACH)
    return TRUE;

  init_logger();

  // place hook on the disassembler
  auto pc64_mod = GetModuleHandleA("pc64.dll");

  // REPLACE THIS WITH THE FUNCTION'S ADDRESS
  static auto constexpr ana_insn_offs = 0x00000000;
  static_assert(ana_insn_offs != 0, "REPLACE THIS WITH THE FUNCTION'S ADDRESS");

  ana_insn = (void *)((uint64)pc64_mod + ana_insn_offs);
  LOG("pc64: %p\nana: %p\n", pc64_mod, ana_insn);
  if (!pc64_mod)
    return FALSE;

  DetourTransactionBegin();
  DetourUpdateThread(GetCurrentThread());
  DetourAttach(&(PVOID &)ana_insn, (PVOID)ana_insn_replacement);
  if (DetourTransactionCommit() == NO_ERROR) {
    LOG("attached hook!\n");
  } else {
    LOG("did not attach hook\n");
  }

  return TRUE;
}
